import json

from google.cloud.bigquery import Client, LoadJobConfig, QueryJobConfig, Table
from google.cloud.exceptions import BadRequest, Conflict

from .exceptions import Invalid
from .logger import DummyCLI
from .query import InsertQuery, LoadQuery, BigQueryUpdateQuery

TIMEOUT_MS = 20000


class BigQuery(object):
    client = Client()

    def __init__(self, cli=None):
        self.cli = cli
        if cli is None:
            self.cli = DummyCLI()

    @property
    def project(self):
        return self.client.project

    def create_table(self, dataset, table_name, schema, recreate=False):
        """Create the table in BigQuery.

        If the table already exists, do nothing.
        Unless recreate is set to True, in which case we drop the table first.
        """
        dataset = self.client.dataset(dataset)
        table_ref = dataset.table(table_name)
        table = Table(table_ref, schema=schema)

        try:
            self.client.create_table(table)
            self.cli.log('Creating "{}" table (BigQuery) ...'.format(table_name))
            self.cli.done()
        except Conflict:
            # Table already exists
            if recreate:
                self.cli.log('Re-creating "{}" table (BigQuery) ...'.format(table_name))
                self.client.delete_table(table)
                self.cli.done()
            else:
                return table

        return table

    def insert(self, table_name, key_field, schema):
        """Run an INSERT query."""
        query = InsertQuery(table_name=table_name, key_field=key_field, schema=schema,
                            prefix='dimensions')
        self.cli.log('Inserting into "{}" table (BigQuery) ...'.format(query.table))

        job_config = QueryJobConfig()
        query_job = self.client.query(query.sql, job_config=job_config)
        query_job.result(timeout=TIMEOUT_MS)

        self.cli.done()

    def load(self, dataset, table_name, schema, payload):
        """Load a payload to BigQuery."""
        self.cli.log('Loading {} records to "{}" table (BigQuery) ...'.format(
            len(payload), table_name))

        query = LoadQuery(payload, schema=schema)

        dataset = self.client.dataset(dataset)
        table_ref = dataset.table(table_name)
        job_config = LoadJobConfig()
        job_config.schema = schema
        job_config.source_format = query.source_format
        job = self.client.load_table_from_file(query.data, table_ref, job_config=job_config)
        try:
            job.result()
        except BadRequest:
            # The error will be set in the job.errors property, so no need to log it
            # again. We do add the requested payload to help troubleshooting.
            job.errors.append({'payload': query.data.read().decode('utf-8')})

        if job.errors:
            raise Invalid(json.dumps(job.errors))

        self.cli.done()

        return table_ref

    def update(self, table_name, key_field, schema):
        """Run an UPDATE query."""
        query = BigQueryUpdateQuery(table_name=table_name, key_field=key_field, schema=schema)
        self.cli.log('Updating "{}" table (BigQuery) ...'.format(query.table))

        job_config = QueryJobConfig()
        query_job = self.client.query(query.sql, job_config=job_config)
        query_job.result(timeout=TIMEOUT_MS)

        self.cli.done()

    def update_table(self, dataset, table_name, schema):
        """Update a table schema in BigQuery.

        BigQuery schema updates only support adding columns, or editing them.
        Removing columns is not supported.
        """
        self.cli.log('Updating "{}" table (BigQuery) ...'.format(table_name))

        dataset = self.client.dataset(dataset)
        table_ref = dataset.table(table_name)
        table = self.client.get_table(table_ref)

        if len(table.schema) <= len(schema):
            table.schema = schema
            self.client.update_table(table, ['schema'])

        self.cli.done()
